import 'dart:collection';

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

abstract class ResolverController<T> {
  final ValueNotifier<T> _stateNotifier;

  ResolverController(this._stateNotifier);

  void run(Stream<T> Function() executor) {
    executor().forEach((element) {
      _stateNotifier.value = element;
    });
  }
}

typedef Predicate<T> = bool Function(T a, T b);

class QueryAspect<T> {
  final Query<T> query;
  final QueryResolver resolver;
  T? _cachedResult;

  QueryAspect(this.query, this.resolver);

  bool shouldNotfiy(ValueNotifier stateNotifier) {
    final newValue = resolver.resolve(stateNotifier, query.params);
    final didChange = query.updateCriteria(_cachedResult, newValue);
    _cachedResult = newValue;

    return didChange;
  }
}

class Query<T> {
  final Object? params;
  final Predicate updateCriteria;

  Query([this.params, Predicate? updateCriteria])
      : this.updateCriteria = updateCriteria ?? _neq;

  T? exec(BuildContext context) {
    final r = context.getResolverForQueryType<T>();
    final container = r.getContainer(context);
    return r.resolve(container._stateNotifier, params);
  }

  T? subscribe(BuildContext context) {
    final r = context.getResolverForQueryType<T>();
    final container = r.getContainer(context);
    final element = context.getContainerHolder(container);

    final result = r.resolve(container._stateNotifier, params);
    final aspect = QueryAspect(this, r);
    aspect._cachedResult = result;

    context.dependOnInheritedElement(element, aspect: aspect);

    return result;
  }

  unsubscribe(Element element) {
    final r = element.getResolverForQueryType<T>();
    final container = r.getContainer(element);
    final el = element.getContainerHolder(container);
    el.unsubscribe(element);
  }

  Widget builder(QueryResultBuilder<T> builder) {
    return QueryBuilder<T>(query: this, builder: builder);
  }
}

typedef QueryCallback<T, K> = K Function(
  ResolverController<T> ctrl,
  T state,
  Object? params,
);

class QueryController<T> extends ResolverController<T> {
  QueryController(ValueNotifier<T> stateNotifier) : super(stateNotifier);
}

abstract class Mutation<T> {
  final T payload;

  Mutation(this.payload);

  void exec(BuildContext context) {
    final r = context.getMutationResolverOfType(runtimeType);
    final container = r.getContainer(context);
    final newState = r.resolve(container._stateNotifier, this);
    container._stateNotifier.value = newState;
    return newState;
  }
}

typedef MutationCallback<T, K> = T Function(
  MutationController<T> ctrl,
  T state,
  K? mutation,
);

class MutationController<T> extends ResolverController<T> {
  MutationController(ValueNotifier<T> stateNotifier) : super(stateNotifier);
}

typedef MutationObserverCallback<T, M, S> = Stream<Mutation> Function(
  T ownState,
  M mutation,
  S oldState,
  S newState,
);

class MutationObserver<T, M extends Mutation, S> {
  final MutationObserverCallback<T, M, S> callback;

  Type resolverType;

  MutationObserver(this.callback)
      : resolverType =
            MutationResolver<S, M>((_, state, __) => state).runtimeType;

  Stream<Mutation> exec(T ownState, M mutation, S oldState, S newState) {
    return callback(ownState, mutation, oldState, newState);
  }
}

typedef StateContainerBoot<T> = Stream<T> Function(BuildContext context);

class StateContainer<T> {
  final ValueNotifier<T> _stateNotifier;

  late final _queryResolvers = Context<Type, QueryResolver>({
    T: QueryResolver<T, T>((ctrl, state, _) => state),
  });
  final _mutationResolvers = Context<Type, MutationResolver>();

  StateContainerBoot? _boot;

  T get _state {
    return _stateNotifier.value;
  }

  set _state(T state) {
    _stateNotifier.value = state;
  }

  Type get stateType => T;

  static StateContainer<T> of<T>(BuildContext context) =>
      context.getContainerOfType<T>();

  StateContainer(T state) : _stateNotifier = ValueNotifier(state);

  void boot(StateContainerBoot<T> boot) {
    _boot = boot;
  }

  void query<K>(QueryCallback<T, K> resolver) {
    final queryResolver = QueryResolver<T, K>(resolver);
    _queryResolvers[queryResolver.keyType] = queryResolver;
  }

  void mutation<K extends Mutation>(MutationCallback<T, K> resolver) {
    final mutationResolver = MutationResolver<T, K>(resolver);
    _mutationResolvers[K] = mutationResolver;
  }

  StateContainer<T> fork() {
    final c = StateContainer<T>(_state);

    c._queryResolvers.addAll(_queryResolvers);
    c._mutationResolvers.addAll(_mutationResolvers);

    return c;
  }
}

extension _ElementsQueries on BuildContext {
  @protected
  _StateContainerElement? getContainerElement() =>
      this.getElementForInheritedWidgetOfExactType<_StateContainer>()
          as _StateContainerElement?;

  StateContainer<T> getContainerOfType<T>() {
    try {
      final el = getContainerElement();
      final container = el?._containers[T];
      return container! as StateContainer<T>;
    } catch (err) {
      throw new Exception('No StateContainer<$T> found');
    }
  }

  Resolver getResolverForType<R>(Type type) {
    final el = getContainerElement();
    Resolver? r;

    if (R == QueryResolver) {
      r = el?._queryResolvers[type];
    } else {
      r = el?._mutationResolvers[type];
    }

    if (r == null) {
      throw new Exception('No resolver found for $type');
    }

    return r;
  }

  QueryResolver getResolverForQueryType<T>() {
    final r = getResolverForType<QueryResolver>(T);
    return r as QueryResolver;
  }

  MutationResolver getMutationResolverOfType(Type type) {
    final r = getResolverForType<MutationResolver>(type);
    return r as MutationResolver;
  }

  _StateContainerElement getContainerHolder(container) {
    final el = getContainerElement();
    return el!._elements[container.runtimeType]!;
  }
}

extension FrameworkOps on BuildContext {
  T? query<T>([Object? params]) {
    return Query<T>(params).exec(this);
  }

  Widget subscribe<T>({
    Object? params,
    Widget? child,
    required Widget Function(BuildContext context, T value, Widget? child)
        builder,
  }) {
    return QueryBuilder<T>(
      query: Query<T>(params),
      builder: builder,
      child: child,
    );
  }

  void mutation<T extends Mutation>(T mutation) {
    mutation.exec(this);
  }
}

extension FrameworkWidgets on Widget {
  Widget withContainer<T>(StateContainer<T> container) {
    return StateContainerProvider(
      container: container,
      child: this,
    );
  }
}

abstract class ContextualizedMutation<T, C> extends Mutation<T> {
  Type get contextType => C;
  ContextualizedMutation(T payload) : super(payload);
}

abstract class Resolver<P, T> {
  T? resolve(ValueNotifier<P> stateNotifier, [dynamic args]);

  Type get contextType => P;
  Type get keyType => T;

  @protected
  StateContainer<P> getContainer(BuildContext context) {
    return context.getContainerOfType<P>();
  }
}

class QueryResolver<P, T> extends Resolver<P, T> {
  final QueryCallback<P, T> _callback;

  QueryResolver(this._callback);

  @override
  T? resolve(stateNotifier, [params]) {
    final ctrl = QueryController<P>(stateNotifier);
    return _callback(ctrl, stateNotifier.value, params);
  }
}

class MutationResolver<P, T extends Mutation> extends Resolver<P, P> {
  final MutationCallback<P, T> _callback;

  MutationResolver(this._callback);

  @override
  P resolve(stateNotifier, [mutation]) {
    final ctrl = MutationController<P>(stateNotifier);
    return _callback(ctrl, stateNotifier.value, mutation);
  }
}

class _StateContainer extends InheritedWidget {
  final StateContainer value;

  _StateContainer({
    Key? key,
    required this.value,
    required Widget child,
  }) : super(key: key, child: child);

  @override
  bool updateShouldNotify(_StateContainer oldWidget) {
    return value._state != oldWidget.value._state;
  }

  @override
  InheritedElement createElement() {
    return _StateContainerElement(this);
  }
}

class _Update {
  final List<Element> dirty = [];
  final List<Element> inactive = [];
}

class Context<T, K> {
  Context? parent;

  final HashMap<T, K> _storage;

  Context([Map<T, K>? initial]) : _storage = HashMap<T, K>.from(initial ?? {});

  inherit(Context<T, K> other) {
    parent = other;
    _storage.addAll(other._storage);
  }

  addAll(Context<T, K> other) {
    _storage.addAll(other._storage);
  }

  K? operator [](T key) {
    return _storage[key];
  }

  operator []=(T key, K value) {
    _storage[key] = value;
  }

  remove(T key) {
    _storage.remove(key);
  }

  clear() {
    parent = null;
    _storage.clear();
  }
}

class _StateContainerElement extends InheritedElement {
  final StateContainer container;

  _StateContainerElement(_StateContainer widget)
      : container = widget.value,
        super(widget);

  final _containers = Context<Type, StateContainer>();
  final _queryResolvers = Context<Type, QueryResolver>();
  final _mutationResolvers = Context<Type, MutationResolver>();
  final _elements = Context<Type, _StateContainerElement>();
  final _dependencyAspects = Context<Element, QueryAspect>();

  Set<Element> _subscribers = {};

  @override
  void mount(Element? parent, newSlot) {
    final pEl = parent?.getContainerElement();

    if (pEl != null) {
      _containers.inherit(pEl._containers);
      _queryResolvers.inherit(pEl._queryResolvers);
      _mutationResolvers.inherit(pEl._mutationResolvers);
      _elements.inherit(pEl._elements);
    }

    _containers[container.stateType] = container;
    _queryResolvers.addAll(container._queryResolvers);
    _mutationResolvers.addAll(container._mutationResolvers);
    _elements[container.runtimeType] = this;

    container._stateNotifier.addListener(onChange);

    container._boot?.call(this).forEach((element) {
      container._state = element;
    });

    super.mount(parent, newSlot);
  }

  @override
  void updateDependencies(Element dependent, Object? aspect) {
    _subscribers.add(dependent);
    _dependencyAspects[dependent] = aspect as QueryAspect;
    setDependencies(dependent, this);
  }

  void unsubscribe(Element element) {
    _subscribers.remove(element);
    _dependencyAspects.remove(element);
  }

  @override
  void unmount() {
    _subscribers.clear();
    _containers.clear();
    _queryResolvers.clear();
    _mutationResolvers.clear();
    _elements.clear();

    container._stateNotifier.removeListener(onChange);

    super.unmount();
  }

  onChange() {
    final u = _subscribers.fold(_Update(), maybeTriggerElementRebuild);

    u.inactive.forEach((element) {
      _subscribers.remove(element);
    });

    u.dirty.forEach((element) {
      element.markNeedsBuild();
    });
  }

  _Update maybeTriggerElementRebuild(_Update update, Element element) {
    if (getDependencies(element) != this) {
      return update..inactive.add(element);
    }

    final aspect = _dependencyAspects[element];

    if (aspect!.shouldNotfiy(container._stateNotifier)) {
      return update..dirty.add(element);
    }

    return update;
  }
}

typedef QueryResultBuilder<T> = Widget Function(
  BuildContext context,
  T data,
  Widget? child,
);

class QueryBuilder<T> extends Widget {
  final Query<T> query;
  final QueryResultBuilder<T> builder;
  final Widget? child;

  QueryBuilder({
    Key? key,
    required this.query,
    required this.builder,
    this.child,
  }) : super(key: key);

  @override
  Element createElement() {
    return QueryBuilderElement<T>(this);
  }
}

class QueryBuilderElement<T> extends ComponentElement {
  QueryBuilderElement(Widget widget) : super(widget);

  @override
  Widget build() {
    final w = widget as QueryBuilder<T>;

    final d = w.query.subscribe(this);
    return w.builder(this, d!, w.child);
  }

  @override
  void deactivate() {
    (widget as QueryBuilder).query.unsubscribe(this);
    super.deactivate();
  }
}

final _neq = (a, b) => a != b;

class _RebuildScope extends StatelessWidget {
  final StateContainer container;
  final Widget child;

  const _RebuildScope({
    Key? key,
    required this.container,
    required this.child,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return _StateContainer(
      value: container,
      child: child,
    );
  }
}

class StateContainerProvider extends _RebuildScope {
  StateContainerProvider({
    required StateContainer container,
    required Widget child,
  }) : super(container: container, child: child);
}

class TextQuery<T> {
  final String text;
  TextQuery(this.text);
}

final Map<Widget, StateContainer> _containersMap = {};

abstract class Binding<T> extends StatelessWidget {
  final Widget child;

  const Binding({Key? key, required this.child}) : super(key: key);

  void init() {}

  @mustCallSuper
  void dispose() {
    _containersMap.remove(this);
  }

  StateContainer<T> create(BuildContext context);

  StateContainer<T> get _container => _containersMap[this] as StateContainer<T>;

  StatelessElement createElement() {
    return StatelessElementWithInitDispose<T>(this);
  }

  StateContainer<T> container<T>(T state) => StateContainer<T>(state);

  update(T newState) {
    _container._state = newState;
  }

  @override
  Widget build(BuildContext context) {
    _containersMap[this] ??= create(context);
    return child.withContainer(_containersMap[this]!);
  }
}

class StatelessElementWithInitDispose<T> extends StatelessElement {
  StatelessElementWithInitDispose(StatelessWidget widget) : super(widget);

  @override
  void mount(Element? parent, Object? newSlot) {
    (widget as Binding).init();
    super.mount(parent, newSlot);
  }

  @override
  void update(covariant StatelessWidget newWidget) {
    final c = _containersMap.remove(widget)!;
    _containersMap[newWidget] = c;

    (widget as Binding).dispose();
    (newWidget as Binding).init();

    super.update(newWidget);
  }

  @override
  void unmount() {
    (widget as Binding).dispose();
    _containersMap.remove(widget);
    super.unmount();
  }
}

class TextEditingControllerBinding<T> extends Binding<String> {
  final TextEditingController controller;

  TextEditingControllerBinding({
    Key? key,
    required Widget child,
    required this.controller,
  }) : super(key: key, child: child);

  @override
  StateContainer<String> create(BuildContext context) {
    return container<String>(controller.text)
      ..query<TextQuery<T>>((ctrl, state, params) => TextQuery<T>(state));
  }

  @override
  void init() {
    controller.addListener(onChange);
    super.init();
  }

  onChange() {
    update(controller.text);
  }

  @override
  void dispose() {
    controller.removeListener(onChange);
    super.dispose();
  }
}

class Boxed<T> {
  final T payload;
  Boxed(this.payload);

  @override
  int get hashCode => payload.hashCode;

  @override
  bool operator ==(Object other) {
    return other.hashCode == hashCode;
  }
}
